#define _GNU_SOURCE

#ifdef WINDOWS
#include <windows.h>
#endif

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <GL/gl.h>

#include "fieldsat.h"

struct {
	const char *tag;
	int type;
} FIELDSAT_TYPES[] = {
	{ "GO", SAMPLE_TYPE_GO },
	{ "S1", SAMPLE_TYPE_S1 },
	{ "S2", SAMPLE_TYPE_S2 },
	{ "SD", SAMPLE_TYPE_SD },
	{ NULL, -1 }
};

double ACC_LOW_ZERO[3] = { 514.98630137, 517.28169014, 513.78219178 };
double ACC_LOW_VARIANCE[3] = { 0.01351098, 0.02922875, 0.17036780 };
/*
double ANG_VEL_ZERO[3] = { 522.15342466, 504.88356164, 518.14794521 };
double ANG_VEL_VARIANCE[3] = { 0.17646087, 0.12479827, 0.14523550 };
*/
double ANG_VEL_ZERO[3] = { 522.16964873, 504.69166018, 517.73302349 };
double ANG_VEL_VARIANCE[3] = { 0.97858638, 0.79441385, 0.41600737 };

const double ACC_LOW_GAIN[3] = { 1.7326314, 1.7326314, 1.7326314 };
const double ANG_VEL_GAIN[3] = { 90.0 / 111.824312, 90.0 / 111.824312, 90.0 / 111.824312 };

#ifdef WINDOWS

HANDLE serial_handle = INVALID_HANDLE_VALUE;

/*
	http://www.robbayer.com/files/serial-win.pdf
*/

static void close_handle(void)
{
	fprintf(stderr, "Closing serial port.\n");
	CloseHandle(serial_handle);
}

static HANDLE serial_open(unsigned int num)
{
	char buf[6];
	HANDLE hSerial;
	DCB dcbSerialParams = { 0 };
	COMMTIMEOUTS timeouts = { 0 };
	
	snprintf(buf, 6, "COM%u", num);
	hSerial = CreateFile(buf, GENERIC_READ, 0, 0,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	if (hSerial == INVALID_HANDLE_VALUE)
		return hSerial;
		
	atexit(close_handle);
	
	dcbSerialParams.DCBlength = sizeof(dcbSerialParams);
	if (!GetCommState(hSerial, &dcbSerialParams)) {
		fprintf(stderr, "Error getting serial parameters.\n");
		exit(1);
	}
	dcbSerialParams.BaudRate = CBR_38400;
	dcbSerialParams.ByteSize = 8;
	dcbSerialParams.StopBits = ONESTOPBIT;
	dcbSerialParams.Parity = NOPARITY;
	
	if (!SetCommState(hSerial, &dcbSerialParams)) {
		fprintf(stderr, "Error setting serial parameters.\n");
		exit(1);
	}
		
	timeouts.ReadIntervalTimeout = 500;
	timeouts.ReadTotalTimeoutConstant = 500;
	timeouts.ReadTotalTimeoutMultiplier = 10;
	
	if (!SetCommTimeouts(hSerial, &timeouts)) {
		fprintf(stderr, "Error setting serial timeouts.\n");
		exit(1);
	}
	
	return hSerial;
}

#endif

FILE *stream_open(int argc, char *argv[])
{
	char buf[1024];
	char *filename;
	FILE *fp;
	unsigned int com_num;

	if (argc < 2) {
#ifdef WINDOWS
		printf("Enter a file name or COM port number (\"1\" for COM1): ");
		fflush(stdout);
		filename = buf;
		fscanf(stdin, "%1024s", filename);
#else
		return stdin;
#endif
	} else {
		filename = argv[1];
	}
	
	fp = fopen(filename, "rb");
	if (fp == NULL) {
		fprintf(stderr, "Can't open the file %s.\n", filename);
#ifdef WINDOWS
		if (sscanf(filename, "%u", &com_num) != 1) {
			printf("Invalid COM port number.\n");
			sleep(3000);
			exit(1);
		}
		serial_handle = serial_open(com_num);
		if (serial_handle == INVALID_HANDLE_VALUE) {
			fprintf(stderr, "Can't open COM%u.\n", com_num);
			sleep(3000);
			exit(1);
		}
#endif
	}

	return fp;
}

static size_t stream_read(char *buf, size_t count, FILE *fp)
{
#ifdef WINDOWS
	DWORD dwBytesRead;
	unsigned int i;
	
	if (fp == NULL) {
		if (serial_handle != INVALID_HANDLE_VALUE) {
			ReadFile(serial_handle, buf, count, &dwBytesRead, NULL);
			return dwBytesRead;
		} else {
			return 0;
		}
	} else {
		return fread(buf, 1, count, fp);
	}
#else
	return fread(buf, 1, count, fp);
#endif
}

void load_config(const char *filename)
{
	double acc_low[3], ang_vel[3];
	FILE *config;
	
	config = fopen(filename, "r");
	if (config == NULL) {
		fprintf(stderr, "No config file found.\n");
		return;
	}
	
	if (fscanf(config, "%lf %lf %lf %lf %lf %lf", &acc_low[0], &acc_low[1], &acc_low[2],
		&ang_vel[0], &ang_vel[1], &ang_vel[2]) != 6) {
		fprintf(stderr, "Error reading configuration data.\n");
	}
	
	fclose(config);
	
	ACC_LOW_ZERO[0] = acc_low[0];
	ACC_LOW_ZERO[1] = acc_low[1];
	ACC_LOW_ZERO[2] = acc_low[2];
	ANG_VEL_ZERO[0] = ang_vel[0];
	ANG_VEL_ZERO[1] = ang_vel[1];
	ANG_VEL_ZERO[2] = ang_vel[2];
}

#define SYNCHRONIZE_BYTES 50

int synchronize(FILE *fp)
{
	char tag[3], buf[32];
	unsigned int i, count;

	count = 0;
	memset(tag, '\0', 3);
	for (;;) {
		tag[0] = tag[1];
		if (stream_read(&tag[1], 1, fp) != 1)
			return 0;
		for (i = 0; FIELDSAT_TYPES[i].tag != NULL; i++) {
			if (strcmp(tag, FIELDSAT_TYPES[i].tag) == 0)
				break;
		}
		if (FIELDSAT_TYPES[i].tag != NULL)
			break;
		count++;
		if (count >= SYNCHRONIZE_BYTES)
			return 0;
	}
	if (stream_read(buf, 32 - 6, fp) != 32 - 6)
		return 0;

	return 1;
}

/* You may have to call synchronize before calling this function. */
int read_sample(FILE *fp, struct fieldsat_sample *sample)
{
	char tag[3], buf[32];
	unsigned int i;

	tag[2] = '\0';

	if (stream_read(buf, 32, fp) != 32)
		return 0;
	memmove(tag, buf + 4, 2);

	for (i = 0; FIELDSAT_TYPES[i].tag != NULL; i++) {
		if (strcmp(tag, FIELDSAT_TYPES[i].tag) == 0)
			break;
	}
	if (FIELDSAT_TYPES[i].tag == NULL)
		return 0;

	sample->timestamp = *((unsigned long *) buf);
	sample->type = FIELDSAT_TYPES[i].type;

	switch (sample->type) {
		case SAMPLE_TYPE_S1:
			sample->s1.acc_low[0] = *((short *) (buf + 6));
			sample->s1.acc_low[1] = *((short *) (buf + 8));
			sample->s1.acc_low[2] = *((short *) (buf + 10));
#ifdef OLD_FORMAT
			/* Skip 6 bytes for nonexistent high-g acceleration. */
			sample->s1.ang_vel[0] = *((short *) (buf + 18));
			sample->s1.ang_vel[1] = *((short *) (buf + 20));
			sample->s1.ang_vel[2] = *((short *) (buf + 22));
			sample->s1.solar_luminosity = *((short *) (buf + 24));
			sample->s1.solar_voltage = *((short *) (buf + 26));
#else
			sample->s1.ang_vel[0] = *((short *) (buf + 12));
			sample->s1.ang_vel[1] = *((short *) (buf + 14));
			sample->s1.ang_vel[2] = *((short *) (buf + 16));
			sample->s1.solar_luminosity = *((short *) (buf + 18));
			sample->s1.solar_voltage = *((short *) (buf + 20));
#endif
			break;
		case SAMPLE_TYPE_S2:
			sample->s2.pressure = *((short *) (buf + 6));
			sample->s2.temp_int = *((short *) (buf + 8));
			sample->s2.temp_ext = *((short *) (buf + 10));
			break;
	}

	return 1;
}

void sample_init(struct fieldsat_sample *sample)
{
	sample->timestamp = 0UL;

	sample->s1.acc_low[0] = 0.0;
	sample->s1.acc_low[1] = 0.0;
	sample->s1.acc_low[2] = 0.0;
	sample->s1.ang_vel[0] = 0.0;
	sample->s1.ang_vel[1] = 0.0;
	sample->s1.ang_vel[2] = 0.0;
	sample->s1.solar_luminosity = 0.0;
	sample->s1.solar_voltage = 0.0;

	sample->s2.pressure = 0.0;
	sample->s2.temp_int = 0.0;
	sample->s2.temp_ext = 0.0;
}

void sample_print(const struct fieldsat_sample *sample)
{
	switch (sample->type) {
		case SAMPLE_TYPE_GO:
			printf("GO\n");
			break;
		case SAMPLE_TYPE_S1:
			printf("X acc low: %f\n", sample->s1.acc_low[0]);
			printf("Y acc low: %f\n", sample->s1.acc_low[1]);
			printf("Z acc low: %f\n", sample->s1.acc_low[2]);
			printf("X ang_vel high: %f\n", sample->s1.ang_vel[0]);
			printf("Y ang_vel high: %f\n", sample->s1.ang_vel[1]);
			printf("Z ang_vel high: %f\n", sample->s1.ang_vel[2]);
			printf("Solar lum.: %f\n", sample->s1.solar_luminosity);
			printf("Solar voltage: %f\n", sample->s1.solar_voltage);
			break;
		case SAMPLE_TYPE_S2:
			printf("Pressure: %f\n", sample->s2.pressure);
			printf("Int. temp: %f\n", sample->s2.temp_int);
			printf("Ext. temp: %f\n", sample->s2.temp_ext);
			break;
	}
}

void state_init(struct fieldsat_state *state)
{
	state->rotation[0] = 1.0;
	state->rotation[1] = state->rotation[2] = state->rotation[3] = 0.0;

	state->pressure = 0.0;

	state->solar_is_deployed = 0;
	state->solar_angle = 10.0;
}

static void quaternion_print(const double *q)
{
	printf("[ %+02.3f; %+02.3f, %+02.3f, %+02.3f ]\n", q[0], q[1], q[2], q[3]);
}

static void quaternion_mult(const double *q1, const double *q2, double *result)
{
	double r[4];

	r[0] = q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2] - q1[3] * q2[3];
	r[1] = q1[0] * q2[1] + q2[0] * q1[1] + q1[2] * q2[3] - q1[3] * q2[2];
	r[2] = q1[0] * q2[2] + q2[0] * q1[2] + q1[3] * q2[1] - q1[1] * q2[3];
	r[3] = q1[0] * q2[3] + q2[0] * q1[3] + q1[1] * q2[2] - q1[2] * q2[1];

	result[0] = r[0];
	result[1] = r[1];
	result[2] = r[2];
	result[3] = r[3];
}

static void vector_rotate(const double *v, const double *q, double *result)
{
	double v_t[4], q_t[4];

	v_t[0] = 0.0;
	v_t[1] = v[0];
	v_t[2] = v[1];
	v_t[3] = v[2];

	q_t[0] = q[0];
	q_t[1] = q[1];
	q_t[2] = q[2];
	q_t[3] = q[3];

	quaternion_mult(q_t, v_t, v_t);
	q_t[1] = -q_t[1];
	q_t[2] = -q_t[2];
	q_t[3] = -q_t[3];
	quaternion_mult(v_t, q_t, v_t);

	result[0] = v_t[1];
	result[1] = v_t[2];
	result[2] = v_t[3];
}

#define SQUARE(x) ((x) * (x))

static void state_update_s1(struct fieldsat_state *state,
	const struct fieldsat_s1 *s1, double t)
{
	double q[4], down[3];
	double down_angle, acc_angle, v_angle, front_angle, angle, len;
	double tmp;
	unsigned int i;

	for (i = 0; i < 3; i++) {
		tmp = (s1->acc_low[i] - ACC_LOW_ZERO[i])
			/ 512.0 * ACC_LOW_GAIN[i] * 5.0 * 9.8;
		/*
		if (tmp > 50.0) {
			if (i == 1)
				tmp = 9.8;
			else
				tmp = 0.0;
		}
		*/
		state->acceleration[i] = state->acceleration[i] + 0.05 * (tmp - state->acceleration[i]);
		state->avg_acc[i] += 0.3 * (tmp - state->avg_acc[i]);
	}
	for (i = 0; i < 3; i++) {
		tmp = -(s1->ang_vel[i] - ANG_VEL_ZERO[i])
			/ 512.0 * ANG_VEL_GAIN[i] * 300.0 * 2.0;
		state->ang_vel[i] = state->ang_vel[i] + 0.2 * (tmp - state->ang_vel[i]);
		/*
		ANG_VEL_ZERO[i] += 0.001 * (s1->ang_vel[i] - ANG_VEL_ZERO[i]);
		*/
	}
#ifdef OLD_FORMAT
	state->solar_luminosity = 1.0;
	state->solar_voltage = 20.0;
#else
	state->solar_luminosity = 1.0 - (s1->solar_luminosity) / 1023.0;
	state->solar_voltage = s1->solar_voltage / 1023.0 * 40.0;
#endif

	/* Rotate about the three axes, one at a time. */
	q[0] = cos(state->ang_vel[0] * t * M_PI_2 / 180.0);
	q[1] = sin(state->ang_vel[0] * t * M_PI_2 / 180.0);
	q[2] = q[3] = 0;
	quaternion_mult(state->rotation, q, state->rotation);
	q[0] = cos(state->ang_vel[1] * t * M_PI_2 / 180.0);
	q[2] = sin(state->ang_vel[1] * t * M_PI_2 / 180.0);
	q[1] = q[3] = 0;
	quaternion_mult(state->rotation, q, state->rotation);
	q[0] = cos(state->ang_vel[2] * t * M_PI_2 / 180.0);
	q[3] = sin(state->ang_vel[2] * t * M_PI_2 / 180.0);
	q[1] = q[2] = 0;
	quaternion_mult(state->rotation, q, state->rotation);

	/* These steps attempt to stabilize the gyroscopes readings with those
	   of the accelerometers. */

	/* Calculate the "down" vector according to the gyros. */
	down[0] = 0.0;
	down[1] = -1.0;
	down[2] = 0.0;
	vector_rotate(down, state->rotation, down);

	/* Find the angle the down vector makes with the vertical. */
	down_angle = acos(down[1] / sqrt(SQUARE(down[0]) + SQUARE(down[1]) + SQUARE(down[2])));
	/* Find the angle the acceleration vector makes with the vertical. */
	acc_angle = acos(-state->acceleration[1] / sqrt(SQUARE(state->acceleration[0]) + SQUARE(state->acceleration[1]) + SQUARE(state->acceleration[2])));
	/* Calculate some fraction of the difference between them. */
	angle = (down_angle - acc_angle);
	angle *= 0.08;
	/* Rotate around a perpendicular axis. */
	len = sqrt(SQUARE(down[0]) + SQUARE(down[2]));
	if (len < 0.001) {
		/* If the projection of the vector is too short, the axis is
		   indeterminate. Establish a preference for rotating straight
		   back. */
		down[2] = 1.0;
		len = sqrt(SQUARE(down[0]) + SQUARE(down[2]));
	}
	q[0] = cos(angle / 2.0);
	q[1] = -down[2] / len * sin(angle / 2.0);
	q[2] = 0.0;
	q[3] = down[0] / len * sin(angle / 2.0);
	quaternion_mult(q, state->rotation, state->rotation);

	/* Find the angle of the forward-pointing vector according to the
	   accelerometers. */
	v_angle = atan2(state->acceleration[0], state->acceleration[2]);
	/* FInd the angle of the forward-pointing vector according to the
	   current rotation. See
	   http://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles;
	   we are calcuating the yaw. */
	front_angle = atan2(2.0 * (state->rotation[0] * state->rotation[3]
		+ state->rotation[1] * state->rotation[2]),
		1.0 - 2.0 * (SQUARE(state->rotation[2]) + SQUARE(state->rotation[3])));
	/* Calculate the difference. */
	angle = (v_angle - front_angle);
	/* This correction is most valid when the package is on its side, so
	   multiply by the sine of the angle from the vertical. */
	angle *= -0.002 * sin(down_angle);
	q[0] = cos(angle / 2.0);
	q[1] = 0.0;
	q[2] = sin(angle / 2.0);
	q[3] = 0.0;
	quaternion_mult(state->rotation, q, state->rotation);
}

static void state_update_s2(struct fieldsat_state *state,
	const struct fieldsat_s2 *s2, double t)
{
	double tmp;

	tmp = s2->pressure / 1023.0 * 100.0;
	if (state->pressure == 0.0)
		state->pressure = tmp;
	else
		state->pressure = state->pressure + 0.002 * (tmp - state->pressure);
	state->altitude = 15500.0 * (5.0 - log10(state->pressure * 1000.0));
	state->temp_int = (s2->temp_int
		/ 1023.0 * 200 - 50.0) * 9.0 / 5.0 + 32.0;
	state->temp_ext = (s2->temp_ext
		/ 1023.0 * 200 - 50.0) * 9.0 / 5.0 + 32.0;
}

static void state_update_sd(struct fieldsat_state *state,
	const struct fieldsat_boolean *boolean, double t)
{
	state->solar_is_deployed = state->solar_is_deployed || boolean->value;
}

void state_update(struct fieldsat_state *state,
	const struct fieldsat_sample *sample, double t)
{
	switch (sample->type) {
		case SAMPLE_TYPE_S1:
			state_update_s1(state, &sample->s1, t);
			break;
		case SAMPLE_TYPE_S2:
			state_update_s2(state, &sample->s2, t);
			break;
		case SAMPLE_TYPE_SD:
			state_update_sd(state, &sample->boolean, t);
			break;
	}
	if (state->solar_is_deployed && state->solar_angle < SOLAR_MAX_ANGLE) {
		state->solar_angle += SOLAR_RATE * t;
		if (state->solar_angle > SOLAR_MAX_ANGLE)
			state->solar_angle = SOLAR_MAX_ANGLE;
	}
}

void state_print(const struct fieldsat_state *state)
{
	printf("Position: (%f, %f, %f)\n", state->position[0], state->position[1], state->position[2]);
	printf("Velocity: (%f, %f, %f)\n", state->velocity[0], state->velocity[1], state->velocity[2]);
	printf("Acceleration: (%f, %f, %f)\n", state->acceleration[0], state->acceleration[1], state->acceleration[2]);
	printf("Rotation: ");
	quaternion_print(state->rotation);
	printf("Pressure: (%f)\n", state->pressure);
	printf("Internal temperature: (%f)\n", state->temp_int);
	printf("External temperature: (%f)\n", state->temp_ext);
}
