#!/usr/bin/env python

"""DemoSat library code."""

import copy
import datetime
import math
import unittest

# The CPU frequency of the package's microcontroller.
F_CPU = 8000000.0

ACCELERATION_SIGN = (1, 1, 1)

ACCELERATION_PERMUTATION = (0, 1, 2)
#ACCELERATION_PERMUTATION = (0, 2, 1)
#ACCELERATION_PERMUTATION = (1, 0, 2)
#ACCELERATION_PERMUTATION = (1, 2, 0)
#ACCELERATION_PERMUTATION = (2, 0, 1)
#ACCELERATION_PERMUTATION = (2, 1, 0)

#ANG_VEL_PERMUTATION = (0, 1, 2)
ANG_VEL_PERMUTATION = (0, 2, 1)
#ANG_VEL_PERMUTATION = (1, 0, 2)
#ANG_VEL_PERMUTATION = (1, 2, 0)
#ANG_VEL_PERMUTATION = (2, 0, 1)
#ANG_VEL_PERMUTATION = (2, 1, 0)

ANG_VEL_SIGN = (1, 1, 1)

class Sample(object):
	"""This is a weighted average sample."""
	def __init__(self, init_value = None, response = 1.0):
		self.value = init_value
		assert 0.0 <= response <= 1.0
		self.response = float(response)
		self.prev_time = None

	def update(self, new_value, time):
		if self.prev_time is None:
			self.prev_time = time
		if self.value is None:
			self.value = new_value
			self.prev_time = time
		elif new_value is not None:
			timedelta = time - self.prev_time
			# print "%.4f: %.4f" % (self.response, (1.0 - (1.0 - self.response) ** timedelta))
			# self.value += (1.0 - (1.0 - self.response ** 0.05) ** (timedelta)) * (new_value - self.value)
			# print "%.5f" % (1.0 - (1.0 - self.response ** 0.05) ** (timedelta))
			self.value += self.response * (new_value - self.value)
			self.prev_time = time

	def is_current(self, time):
		return self.prev_time >= time

class Quaternion(object):
	def __init__(self, a = 1.0, b = 0.0, c = 0.0, d = 0.0):
		self.a = a
		self.b = b
		self.c = c
		self.d = d

	def __mul__(self, other):
		result = Quaternion()
		result.a = self.a * other.a - self.b * other.b - self.c * other.c - self.d * other.d
		result.b = self.a * other.b + self.b * other.a + self.c * other.d - self.d * other.c
		result.c = self.a * other.c + self.c * other.a + self.d * other.b - self.b * other.d
		result.d = self.a * other.d + self.d * other.a + self.b * other.c - self.c * other.b
		return result

	def __invert__(self):
		return Quaternion(self.a, -self.b, -self.c, -self.d)

	def normalize(self):
		length = math.sqrt(sum(x * x for x in (self.a, self.b, self.c, self.d)))
		self.a /= length
		self.b /= length
		self.c /= length
		self.d /= length

	def to_matrix(self):
		a, b, c, d = self.a, self.b, self.c, self.d
		return ( a * a + b * b - c * c - d * d, 2 * a * d + 2 * b * c, 2 * b * d - 2 * a * c, 0.0,
			 2 * b * c - 2 * a * d, a * a - b * b + c * c - d * d, 2 * a * b + 2 * c * d, 0.0,
			 2 * a * c + 2 * b * d, 2 * c * d - 2 * a * b, a * a - b * b - c * c + d * d, 0.0,
			 0.0, 0.0, 0.0, 1.0 )

class GPSChecksumError(Exception):
	pass

class GPSTypeError(Exception):
	pass

def gps_calculate_checksum(data):
	sum = 0
	for c in data:
		sum ^= ord(c)
	return sum

def gps_parse_time(time_str):
	"""Return a time from an NMEA time string (hhmmss.sss)."""
	if not time_str:
		return None
	hour = int(time_str[0:2])
	minute = int(time_str[2:4])
	second = float(time_str[4:])
	subsecond = second % 1.0
	second = int(second)
	assert 0.0 <= subsecond < 1.0
	return datetime.time(hour, minute, second, int(subsecond * 1000000))

def gps_parse_date(date_str):
	"""Return a date from an NMEA date string (ddmmyy)."""
	day = int(date_str[0:2])
	month = int(date_str[2:4])
	year = int(date_str[4:])
	year += 2000 # Y3K issues.
	return datetime.date(year, month, day)

def gps_parse_latitude(coord_str, indicator):
	"""Return a numeric latitude from an NMEA coordinate string (ddmm.mmmm)
	and a north-south indicator."""
	if not coord_str or not indicator:
		return None
	degrees = int(coord_str[0:2])
	minutes = float(coord_str[2:])
	latitude = degrees + minutes / 60
	assert indicator == "N" or indicator == "S"
	if indicator == "S":
		latitude = latitude
	return latitude

def gps_parse_longitude(coord_str, indicator):
	"""Return a numeric longitude from an NMEA coordinate string
	(dddmm.mmmm) and an east-west indicator."""
	if not coord_str or not indicator:
		return None
	degrees = int(coord_str[0:3])
	minutes = float(coord_str[3:])
	longitude = degrees + minutes / 60
	assert indicator == "E" or indicator == "W"
	if indicator == "W":
		longitude = -longitude
	return longitude

def gps_parse_altitude(altitude_str, units):
	"""Return an altitude in meters from an NMEA altitude string and unit
	specifier."""
	if not altitude_str or not units:
		return None
	assert units == "M"
	return float(altitude_str)

def vector_normalize(v):
	length = math.sqrt(sum(a * a for a in v))
	return [ a / length for a in v ]

def vector_rotate(v, q):
	assert len(v) == 3
	r = q * Quaternion(0.0, v[0], v[1], v[2]) * ~q
	# assert r.a < 0.00001
	return [r.b, r.c, r.d]

class demosat_state(object):
	"""This class represents the state of a DemoSat package at one instant.
	"""

	# The response of GPS data (latitude, longitude, altitude). Pretty
	# stiff because we trust GPS.
	GPS_RESPONSE = 0.4

	ACCELERATION_ZERO = ( 506.207530, 536.775127, 503.875168 )
	ACCELERATION_GAIN = ( 1.0, 1.91, 1.0 )
	ACCELERATION_RESPONSE = 0.3
	AVG_ACCELERATION_RESPONSE = 0.005

	VELOCITY_DAMPING = 0.05

	ANG_VEL_ZERO = ( 526.747375, 494.418635, 514.748031 )
	ANG_VEL_GAIN = ( 1.0, 1.0, 1.0 )
	ANG_VEL_GAIN = ( 0.5, 0.5, 0.5 )
	ANG_VEL_RESPONSE = 0.2

	DOWN_ANGLE_CORRECTION = 0.005
	FRONT_ANGLE_CORRECTION = 0.005

	def __init__(self, other = None):
		"""Create a new state with default values. If other is not
		None, create a new state that is a copy of other.
		"""
		if other is None:
			# The cumulative CPU counter.
			self.timestamp = None
			self.seconds_elapsed = None
			# The time difference between this state and the
			# previous one, in seconds.
			self.timedelta = 0.0
			self.date = None
			self.time = None
			self.acceleration = (
				Sample(None, self.ACCELERATION_RESPONSE),
				Sample(None, self.ACCELERATION_RESPONSE),
				Sample(None, self.ACCELERATION_RESPONSE)
			)
			self.avg_acceleration = (
				Sample(None, self.AVG_ACCELERATION_RESPONSE),
				Sample(None, self.AVG_ACCELERATION_RESPONSE),
				Sample(None, self.AVG_ACCELERATION_RESPONSE)
			)
			self.velocity = [0.0, 0.0, 0.0]
			self.position = [0.0, 0.0, 0.0]
			self.angular_velocity = (
				Sample(None, self.ANG_VEL_RESPONSE),
				Sample(None, self.ANG_VEL_RESPONSE),
				Sample(None, self.ANG_VEL_RESPONSE)
			)
			self.rotation = Quaternion()
			self.latitude = Sample(None, self.GPS_RESPONSE)
			self.longitude = Sample(None, self.GPS_RESPONSE)
			self.altitude = Sample(None, self.GPS_RESPONSE)
			self.deployment_status = "CLOSED"
			self.num_photos = 0
			self.sample = None
		else:
			self.timestamp = other.timestamp
			self.seconds_elapsed = other.seconds_elapsed
			self.timedelta = other.timedelta
			self.date = copy.copy(other.date)
			self.time = copy.copy(other.time)
			self.acceleration = [copy.copy(a) for a in other.acceleration]
			self.avg_acceleration = [copy.copy(a) for a in other.avg_acceleration]
			self.velocity = copy.copy(other.velocity)
			self.position = copy.copy(other.position)
			self.angular_velocity = [copy.copy(a) for a in other.angular_velocity]
			self.rotation = copy.copy(other.rotation)
			self.latitude = copy.copy(other.latitude)
			self.longitude = copy.copy(other.longitude)
			self.altitude = copy.copy(other.altitude)
			self.deployment_status = other.deployment_status
			self.num_photos = other.num_photos
			self.sample = copy.deepcopy(other.sample)
		self.down = None

	def update(self, sample):
		"""Update the state from a sample as returned by get_samples.
		"""
		# We update the timestamp first so the state's timestamp will
		# be set.
		self.sample = sample
		self.update_pair(("TS", sample["TS"]))
		for pair in sample.items():
			# TS was already updated.
			if pair[0] == "TS":
				continue
			self.update_pair(pair)

		self.position = [p + rv * self.timedelta for (p, rv) in zip(self.position, vector_rotate(self.velocity, self.rotation))]

	def update_unknown(self, name, value):
		raise KeyError("Unknown sample %s=%s" % (name, value))

	def update_pair(self, pair):
		name, value = pair
		handler = demosat_state.update_handlers.get(name, demosat_state.update_unknown)
		handler(self, name, value)

	def update_ts(self, name, value):
		timestamp = int(value, 16)
		if self.timestamp is None:
			self.timestamp = timestamp
		diff = (timestamp - self.timestamp) % (1 << 32)
		if diff < 0:
			diff += 1 << 32
		assert(diff < 1 << 31)
		self.timestamp += diff
		if self.seconds_elapsed is None:
			self.seconds_elapsed = self.timestamp / F_CPU
			prev_seconds_elapsed = self.seconds_elapsed
		else:
			prev_seconds_elapsed = self.seconds_elapsed
			self.seconds_elapsed = self.timestamp / F_CPU
		self.timedelta = self.seconds_elapsed - prev_seconds_elapsed

	def update_deploy(self, name, value):
		self.deployment_status = value

	def update_numphotos(self, name, value):
		self.num_photos = int(value, 16)

	def update_a(self, name, value):
		assert len(value) == 4
		index = ["AX", "AY", "AZ"].index(name)
		acc = int(value, 16)
		# 15-g accelerometers.
		tmp = (acc - self.ACCELERATION_ZERO[index]) / 512.0 * self.ACCELERATION_GAIN[index] * 9.8 * 15.0
		tmp *= ACCELERATION_SIGN[index]
		index = ACCELERATION_PERMUTATION[index]
		self.avg_acceleration[index].update(tmp, self.seconds_elapsed)
		self.acceleration[index].update(tmp, self.seconds_elapsed)

		self.velocity[index] += (self.avg_acceleration[index].value - self.acceleration[index].value) * self.timedelta

	def update_r(self, name, value):
		AXES = ( (1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0) )
		assert len(value) == 4
		index = ["RX", "RY", "RZ"].index(name)
		ang_vel = int(value, 16)
		# 300 degree-per-second gyros.
		# print (ang_vel - self.ANG_VEL_ZERO[index])
		tmp = (ang_vel - self.ANG_VEL_ZERO[index]) / 512.0 * self.ANG_VEL_GAIN[index] * 300.0
		assert(-350 < tmp < 350)
		tmp *= ANG_VEL_SIGN[index]
		index = ANG_VEL_PERMUTATION[index]
		self.angular_velocity[index].update(tmp, self.seconds_elapsed)
		angle = self.angular_velocity[index].value * self.timedelta
		# print value, tmp, angle
		self.rotate(angle * math.pi / 180.0, AXES[index])

	def stabilize(self):
		for measurement in (self.avg_acceleration[0].value,
		                    self.avg_acceleration[1].value,
		                    self.avg_acceleration[2].value):
			if measurement is None:
				return

		# These steps attempt to stabilize the gyroscopes' readings with
		# those of the accelerometers.

		# Calculate the "down" vector according to the gyros.
		down = (0.0, -1.0, 0.0)
		down = vector_rotate(down, self.rotation)
		self.down = down

		# Find the angle the down vector makes with the vertical.
		down_angle = math.acos(down[1])

		# Find the angle the acceleration vector makes with the
		# vertical.
		norm_avg_acc = vector_normalize([a.value for a in self.avg_acceleration])
		acc_angle = math.acos(norm_avg_acc[1])
		# print "%8.5f %8.5f" % (down_angle * 180.0 / math.pi, acc_angle * 180.0 / math.pi)

		# Calculate some fraction of the difference between them.
		angle = (acc_angle - down_angle) * self.DOWN_ANGLE_CORRECTION

		# Rotate around a perpendicular axis.
		len = math.sqrt(down[0] * down[0] + down[2] * down[2])
		if len < 0.001:
			# If the projection of the vector is too short, the
			# axis is indeterminate. Establish a preference for
			# rotating straight back.
			down[2] = 1.0
			len = math.sqrt(down[0] * down[0] + down[2] * down[2])
		# q, r
		self.rotate_abs(angle, (down[2], 0.0, -down[0]))
		# print angle / self.DOWN_ANGLE_CORRECTION * 180.0 / math.pi

		# Find the angle of the forward-pointing vector according to
		# the accelerometers.
		v_angle = math.atan2(self.avg_acceleration[0].value, self.avg_acceleration[2].value)

		# Find the angle of the forward-pointing vector according to
		# the current rotation. See
		# http://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles;
		# we are calcuating the yaw.
		front_angle = math.atan2(2.0 * (self.rotation.a * self.rotation.d + self.rotation.b * self.rotation.c),
			1.0 - 2.0 * (self.rotation.c * self.rotation.c + self.rotation.d * self.rotation.d))

		# print [x.value for x in self.avg_acceleration]
		# print "%.3f %.3f %.3f" % (front_angle * 180.0 / math.pi, v_angle * 180.0 / math.pi, math.sin(down_angle))

		# Calculate the difference. This correction is most valid when
		# the package is on its side, so multiply by the sine of the
		# angle from the vertical.
		angle = -(v_angle - front_angle) * self.FRONT_ANGLE_CORRECTION * math.sin(down_angle)
		# r, q
		self.rotate(angle, (0.0, 1.0, 0.0))

		self.velocity = [v * (1.0 - self.VELOCITY_DAMPING) for v in self.velocity]

	def update_gps(self, name, value):
		date = None
		time = None
		latitude = None
		longitude = None
		altitude = None

		data, checksum = value.rsplit("*", 1)
		checksum = int(checksum, 16)
		if data.startswith("$"):
			data = data[1:]
		calculated_checksum = gps_calculate_checksum(data)
		if calculated_checksum != checksum:
			raise GPSChecksumError("%d != %d" % (calculated_checksum, checksum))

		type, data = data.split(",", 1)
		assert(type.startswith("GP"))
		type = type[2:]

		entries = data.split(",")

		if type == "GGA": # Fixed data.
			time = gps_parse_time(entries[0])
			latitude = gps_parse_latitude(entries[1], entries[2])
			longitude = gps_parse_longitude(entries[3], entries[4])
			altitude = gps_parse_altitude(entries[8], entries[9])
		elif type == "GSA": # DOP and active satellites.
			# Nothing for us here.
			pass
		elif type == "GSV": # Satellites in view.
			# Nothing for us here.
			pass
		elif type == "RMC": # Recommended minimum specific data.
			time = gps_parse_time(entries[0])
			latitude = gps_parse_latitude(entries[2], entries[3])
			longitude = gps_parse_longitude(entries[4], entries[5])
			date = gps_parse_date(entries[8])
		else:
			raise GPSTypeError("Unknown type %s" % type)

		if time is not None:
			self.time = time
		if date is not None:
			self.date = date
		self.latitude.update(latitude, self.seconds_elapsed)
		self.longitude.update(longitude, self.seconds_elapsed)
		self.altitude.update(altitude, self.seconds_elapsed)

	def update_photo(self, name, value):
		self.num_photos += 1

	update_handlers = {
		"TS": update_ts,
		"DEPLOY": update_deploy,
		"NUMPHOTOS": update_numphotos,
		"RX": update_r,
		"RY": update_r,
		"RZ": update_r,
		"AX": update_a,
		"AY": update_a,
		"AZ": update_a,
		"GPS": update_gps,
		"PHOTO": update_photo
	}

	def format_date(self):
		DATE_FORMAT = "%Y-%m-%d"
		if self.date is None:
			return None
		return self.date.strftime(DATE_FORMAT)

	def format_time(self):
		TIME_FORMAT = "%H:%M:%S"
		if self.time is None:
			return None
		subsecond_str = "%.3f" % (self.time.microsecond / 1000000.0)
		subsecond_str = subsecond_str[subsecond_str.index("."):]
		return self.time.strftime(TIME_FORMAT) + subsecond_str

	def format_date_time(self):
		DATE_FORMAT = "%Y-%m-%d"
		TIME_FORMAT = "%H:%M:%S"
		date_str = self.format_date()
		if date_str is None:
			date_str = "<no date>"
		time_str = self.format_time()
		if time_str is None:
			time_str = "<no time>"
		return "%s %s" % (date_str, time_str)

	def rotate(self, angle, axis):
		assert len(axis) == 3

		axis = vector_normalize(axis)
		v = [ math.sin(angle / 2.0) * a for a in axis ]
		q = Quaternion(math.cos(angle / 2.0), v[0], v[1], v[2])
		self.rotation = self.rotation * q
		self.rotation.normalize()

	# Rotate around an absolute, not relative, axis. This just means the
	# quaternion is multiplied on the left.
	def rotate_abs(self, angle, axis):
		assert len(axis) == 3

		axis = vector_normalize(axis)
		v = [ math.sin(angle / 2.0) * a for a in axis ]
		q = Quaternion(math.cos(angle / 2.0), v[0], v[1], v[2])
		self.rotation = q * self.rotation
		self.rotation.normalize()

	def get_rotation_matrix(self):
		return self.rotation.to_matrix()

	def __str__(self):
		s = "Seconds elapsed: %.2f\n" % self.seconds_elapsed
		s += "Timestamp: %08X  Date: %s\n" \
			% (self.timestamp, self.format_date_time())
		s += "Coordinates: %s %s %s\n" \
			% (self.latitude.value and "%.8f" % self.latitude.value,
			   self.longitude.value and "%.8f" % self.longitude.value,
			   self.altitude.value and "%4.3f" % self.altitude.value)
		s += "Position: %6.3f %6.3f %6.3f\n" % tuple(self.position)
		s += "Velocity: %6.3f %6.3f %6.3f\n" % tuple(self.velocity)
		s += "Acceleration: %s %s %s\n" \
			% tuple(self.acceleration[i].value and "%6.3f" % self.acceleration[i].value for i in range(0, 3))
		s += "Angular velocity: %s %s %s\n" \
			% tuple(self.angular_velocity[i].value and "%6.3f" % self.angular_velocity[i].value for i in range(0, 3))
		s += "Deploy: %s  Num photos: %d\n" \
			% (self.deployment_status, self.num_photos)
		return s

def get_samples(file):
	"""A generator that returns samples (dicts of pairs in between and
	including a "TS" pair) from a file.
	"""
	pair_gen = get_pairs(file)
	# Look for the first "TS" sample.
	count = 0
	for pair in pair_gen:
		if pair[0] == "TS":
			break
		count += 1
	else:
		raise StopIteration
	if count > 0:
		print "Ignored %d samples looking for a \"TS\"." % count
	d = { "TS": pair[1] }
	for pair in pair_gen:
		if pair[0] == "TS":
			yield d
			d = {}
		assert pair[0] not in d
		d[pair[0]] = pair[1]

def get_pairs(file):
	"""A generator that returns NAME=VALUE pairs from a file.
	"""
	c = file.read(1)
	while True:
		# Skip white space.
		while len(c) > 0 and c in " \n":
			c = file.read(1)
		if c == "" or c == "\0":
			raise StopIteration
		elif c == "#":
			while c != "\n":
				c = file.read(1)
			c = file.read(1)
			continue
		s = []
		while c not in "\0 \n#":
			s.append(c)
			c = file.read(1)
		s = "".join(s)
		pair = s.split("=", 1)
		if len(pair) != 2:
			print "Malformed pair: %s" % repr(s)
			continue
		yield pair

def get_states(file):
	state = demosat_state()
	for sample in get_samples(file):
		prev_state = state
		state = demosat_state(prev_state)
		try:
			state.update(sample)
			state.stabilize()
		except GPSChecksumError:
			print "GPS checksum error."
		else:
			yield state

# Unit tests follow.

class Sample_test(unittest.TestCase):
	def test_value(self):
		s = Sample(10)
		self.assertEqual(s.value, 10)

	def test_response(self):
		s = Sample(10, 0.0)
		s.update(15, 1.0)
		self.assertEqual(s.value, 10)
		s = Sample(10, 1.0)
		s.update(15, 1.0)
		self.assertEqual(s.value, 15)
		s = Sample(10, 0.5)
		s.update(15, 1.0)
		self.assertNotEqual(s.value, 10)
		self.assertNotEqual(s.value, 15)

if __name__ == "__main__":
	unittest.main()
