#!/usr/bin/env python

try:
    import psyco
    psyco.full()
except ImportError:
    pass

import getopt
import math
import random
import sys

from OpenGL.GL import *
from OpenGL.GLU import *
import pygame
from pygame.locals import *

import demosat

DEFAULT_SCREEN_WIDTH = 300
DEFAULT_SCREEN_HEIGHT = 300

DEFAULT_MOVIE_SCREEN_WIDTH = 720
DEFAULT_MOVIE_SCREEN_HEIGHT = 480
DEFAULT_MOVIE_SAMPLE_RATE = 29.97

EDGE_LENGTH = 0.25
THICKNESS = 0.02

CLEAR_COLOR = tuple(c / 255.0 for c in ( 176, 224, 230, 255 ))
COLOR_BLACK = tuple(c / 255.0 for c in ( 24, 24, 24 ))
COLOR_RED = tuple(c / 255.0 for c in ( 255, 0, 0 ))
COLOR_BLUE = tuple(c / 255.0 for c in ( 0, 0, 255 ))
COLOR_YELLOW = tuple(c / 255.0 for c in ( 255, 255, 0 ))
COLOR_SILVER = tuple(c / 255.0 for c in ( 224, 224, 224 ))

PARTICLE_COLOR = tuple(c / 255.0 for c in (24, 24, 24, 96))

TEXT_COLOR = (0, 0, 0)

GLOBAL_AMBIENT = [ 0.2, 0.2, 0.2, 0.2 ]
AMBIENT = [ 0.4, 0.4, 0.4, 0.4 ]
DIFFUSE = [ 0.4, 0.4, 0.4, 0.4 ]
SPECULAR = [ 0.2, 0.2, 0.2, 0.2 ]
POSITION = [ -1.0, 0.2, 0.3, 0.0 ]

screen_width = None
screen_height = None
sample_rate = None

start_time = 0.0
end_time = None
current_time = None
state = demosat.demosat_state()
camera_position = [0.0, 0.0, 0.0]
movie_mode = False

big_font = None

particles = [None] * 50

def init():
    global big_font

    random.seed(0)

    glClearColor(*CLEAR_COLOR)

    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glFrustum(-0.1, 0.1, -0.1, 0.1, 1.0, 100.0)

    glEnable(GL_DEPTH_TEST)

    glEnable(GL_LIGHTING)
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, GLOBAL_AMBIENT)
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE)
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE)

    glLightfv(GL_LIGHT0, GL_AMBIENT, AMBIENT)
    glLightfv(GL_LIGHT0, GL_DIFFUSE, DIFFUSE)
    glLightfv(GL_LIGHT0, GL_SPECULAR, SPECULAR)
    glLightfv(GL_LIGHT0, GL_POSITION, POSITION)
    glEnable(GL_LIGHT0)

    glEnable(GL_COLOR_MATERIAL)

    glEnable(GL_NORMALIZE)

    glEnable(GL_POINT_SMOOTH)

    big_font = pygame.font.Font(pygame.font.get_default_font(), 32)

def update_camera_position(camera_position, state):
    diff = [sp - cp for cp, sp in zip(camera_position, state.position)]
    length = math.sqrt(sum(x * x for x in diff))
    if length == 0.0:
        diff_norm = diff
    else:
        diff_norm = [x / length for x in diff]
    speed = min(length / state.timedelta, 1.5)
    diff = [x * speed * state.timedelta for x in diff_norm]
    for i in range(0, len(camera_position)):
        camera_position[i] += diff[i]
    return camera_position

def update(states):
    global current_time, state, camera_position, particles

    if current_time is None:
        state = states.next()
        current_time = state.seconds_elapsed
    while state.seconds_elapsed < current_time:
        state = states.next()
        camera_position = update_camera_position(camera_position, state)

    for i in range(0, len(particles)):
        if particles[i] is None or max(abs(particles[i][j] - camera_position[j]) for j in range(0, 3)) > 0.5:
            particles[i] = [random.random() * 1.0 - 0.5 + camera_position[j] for j in range(0, 3)]

def degrees(r):
        return r * 180.0 / math.pi

def radians(d):
        return d * math.pi / 180.0

PANEL_CLOSED_ANGLE = 180.0 - degrees(math.acos(1.0 / 3.0))

def model_sat():
    points = (
        ( 0.0, 0.0, -EDGE_LENGTH * math.sin(radians(60)) ),
        ( -EDGE_LENGTH / 2.0, 0.0, 0.0 ),
        ( EDGE_LENGTH / 2.0, 0.0, 0.0),
        ( 0.0, -THICKNESS, -EDGE_LENGTH * math.sin(radians(60)) ),
        ( -EDGE_LENGTH / 2.0, -THICKNESS, 0.0 ),
        ( EDGE_LENGTH / 2.0, -THICKNESS, 0.0)
    )
    def model_panel(color):
        glBegin(GL_TRIANGLES)
        # Interior face.
        glColor3dv(COLOR_BLACK)
        glNormal3d(0.0, 1.0, 0.0)
        glVertex3dv(points[0])
        glVertex3dv(points[1])
        glVertex3dv(points[2])
        # Exterior face.
        glColor3dv(color)
        glNormal3d(0.0, -1.0, 0.0)
        glVertex3dv(points[5])
        glVertex3dv(points[4])
        glVertex3dv(points[3])
        glEnd()
        # Edges.
        glBegin(GL_QUADS)
        glNormal3d(math.cos(radians(60 + 90)), 0, -math.sin(radians(60 + 90)))
        glVertex3dv(points[0])
        glVertex3dv(points[3])
        glVertex3dv(points[4])
        glVertex3dv(points[1])
        glNormal3d(0.0, 0.0, 1.0)
        glVertex3dv(points[1])
        glVertex3dv(points[4])
        glVertex3dv(points[5])
        glVertex3dv(points[2])
        glNormal3d(math.cos(radians(60)), 0, -math.sin(radians(60)))
        glVertex3dv(points[2])
        glVertex3dv(points[5])
        glVertex3dv(points[3])
        glVertex3dv(points[0])
        glEnd()

    def model_rod(color):
        glBegin(GL_LINES)
        glColor3dv(color)
        glVertex3d(points[3][0] + EDGE_LENGTH / 6.0, points[3][1] - THICKNESS / 2.0, points[3][2])
        glVertex3d(points[4][0], points[4][1] - THICKNESS / 2.0, points[4][2] + EDGE_LENGTH / 6.0)
        glEnd()
        glPointSize(3.0)
        glBegin(GL_POINTS)
        glColor3dv(color)
        glVertex3d(points[3][0] + EDGE_LENGTH / 6.0, points[3][1] - THICKNESS / 2.0, points[3][2])
        glEnd()

    open_amount = 0.0

    glPushMatrix()

    glTranslated(0.0,
        -EDGE_LENGTH * math.sin(radians(60)) / 3.0,
        EDGE_LENGTH * math.sin(radians(60)) / 3.0)

    model_panel(COLOR_RED)
    model_rod(COLOR_SILVER)

    glPushMatrix()
    glTranslated(0.25 * EDGE_LENGTH, 0.0, -0.5 * EDGE_LENGTH * math.sin(radians(60)))
    glRotated(-60.0, 0.0, 1.0, 0.0)
    glRotated((1.0 - open_amount) * PANEL_CLOSED_ANGLE, 1.0, 0.0, 0.0)
    model_panel(COLOR_BLUE)
    glPopMatrix()

    glPushMatrix()
    glTranslated(-0.25 * EDGE_LENGTH, 0.0, -0.5 * EDGE_LENGTH * math.sin(radians(60)))
    glRotated(60.0, 0.0, 1.0, 0.0)
    glRotated((1.0 - open_amount) * PANEL_CLOSED_ANGLE, 1.0, 0.0, 0.0)
    model_panel(COLOR_RED)

    glTranslated(-0.25 * EDGE_LENGTH, 0.0, -0.5 * EDGE_LENGTH * math.sin(radians(60)))
    glRotated(60.0, 0.0, 1.0, 0.0)
    glRotated((1.0 - open_amount) * PANEL_CLOSED_ANGLE, 1.0, 0.0, 0.0)
    model_panel(COLOR_BLUE)

    glPopMatrix()

    glPopMatrix()

def format_seconds(n):
    result = "%06.3f" % (n % 60)
    n = int(n) / 60
    result = ("%02d" % (n % 60)) + ":" + result
    n /= 60
    result = ("%d" % (n % 60)) + ":" + result
    return result

def display():
    global state, camera_position, big_font, movie_mode, particles

    glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT)

    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()

    ratio = float(screen_width) / screen_height
    glFrustum(-0.1 * ratio, 0.1 * ratio, -0.1, 0.1, 1.0, 1000.0)

    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()
    # gluLookAt(0.0, 0.0, 3.0, state.position[0], 0.0, state.position[2], 0.0, 1.0, 0.0)
    # gluLookAt(0.0, 0.0, 3.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0)
    glTranslated(0.0, 0.0, -3.0)

    glTranslated(*[-x for x in camera_position])

    glPushMatrix()

    glTranslated(*state.position)

    glMultMatrixd(state.get_rotation_matrix())

    model_sat()

    glPopMatrix()

    # Draw particles.
    glPushAttrib(GL_ENABLE_BIT)
    glEnable(GL_BLEND)
    glDisable(GL_LIGHTING)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
    glColor4dv(PARTICLE_COLOR)
    glPointSize(4.0)
    glBegin(GL_POINTS)
    for p in particles:
        glVertex3dv(p)
    glEnd()
    glPopAttrib()

    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()

    ratio = float(screen_width) / screen_height
    glOrtho(0.0, ratio, 0.0, 1.0, -10.0, 10.0)

    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()

    time_str = state.format_time()
    if time_str is not None:
        draw_text("UTC: " + time_str, big_font, 0.2 * ratio, 0.1)
    time_str = format_seconds(state.seconds_elapsed)
    if time_str is not None:
        draw_text("Mission time: " + time_str, big_font, 0.2 * ratio, 0.2)

    pygame.display.flip()

    if movie_mode:
        pixels = glReadPixelsub(0, 0, screen_width, screen_height, GL_RGB)

        # print len(pixels), len(pixels[0]), len(pixels[0][0])
        # print type(pixels)
        # sys.exit(1)

        pixels = pixels.tostring("C")

        for j in range(screen_height - 1, -1, -1):
            sys.stdout.write(pixels[screen_width * 3 * j:screen_width * 3 * (j + 1)])

def draw_text(str, font, left, bottom):
    text = font.render(str, True, TEXT_COLOR)
    scaled_text = pygame.Surface((512, 512), 0, text)
    scaled_text.blit(text, (0, 512 - text.get_height()))
    # pygame.transform.scale(text, (256, 256))
    
    glEnable(GL_BLEND)
    glDisable(GL_DEPTH_TEST)
    glEnable(GL_TEXTURE_2D)
    glDisable(GL_LIGHTING)

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, 4, scaled_text.get_width(), scaled_text.get_height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, pygame.image.tostring(scaled_text, "RGBA", 1))

    glColor4d(1.0, 1.0, 1.0, 1.0)
    glBegin(GL_QUADS)
    glTexCoord2d(0.0, 0.0)
    glVertex2d(left, bottom)
    glTexCoord2d(1.0, 0.0)
    glVertex2d(left + scaled_text.get_width() / 500.0, bottom)
    glTexCoord2d(1.0, 1.0)
    glVertex2d(left + scaled_text.get_width() / 500.0, bottom + scaled_text.get_height() / 500.0)
    glTexCoord2d(0.0, 1.0)
    glVertex2d(left, bottom + scaled_text.get_height() / 500.0)
    glEnd()

    glEnable(GL_LIGHTING)
    glDisable(GL_TEXTURE_2D)
    glEnable(GL_DEPTH_TEST)
    glDisable(GL_BLEND)

quit = False
time_multiplier = 1.0

def handle_key(key, type):
    global time_multiplier, state, quit

    if type == KEYDOWN:
        if key == K_EQUALS:
            if time_multiplier == -1.0:
                time_multiplier = 0.0
            elif time_multiplier == 0.0:
                time_multiplier = 1.0
            elif time_multiplier < 0.0:
                time_multiplier /= 2.0
            else:
                time_multiplier *= 2.0
        elif key == K_MINUS:
            if time_multiplier == 1.0:
                time_multiplier = 0.0
            elif time_multiplier == 0.0:
                time_multiplier = -1.0
            elif time_multiplier > 0.0:
                time_multiplier /= 2.0
            else:
                time_multiplier *= 2.0
        elif key == K_SPACE:
            print >> sys.stderr, "Current time:", current_time
            print >> sys.stderr, state
        elif key == K_ESCAPE:
            quit = True
    elif type == KEYUP:
        pass

def run(file):
    global start_time, end_time, current_time, quit, movie_mode
    global screen_width, screen_height, sample_rate

    states = demosat.get_states(file)

    if screen_width is None:
        if movie_mode:
            screen_width = DEFAULT_MOVIE_SCREEN_WIDTH
        else:
            screen_width = DEFAULT_SCREEN_WIDTH
    if screen_height is None:
        if movie_mode:
            screen_height = DEFAULT_MOVIE_SCREEN_HEIGHT
        else:
            screen_height = DEFAULT_SCREEN_HEIGHT

    if movie_mode and sample_rate is None:
        sample_rate = DEFAULT_MOVIE_SAMPLE_RATE

    pygame.init()
    surface = pygame.display.set_mode((screen_width, screen_height),
        OPENGL | DOUBLEBUF)

    init()

    if start_time > 0.0:
        print >> sys.stderr, "Skipping to %f seconds." % start_time
    ticks = pygame.time.get_ticks()
    while not quit:
        ticks_delta = pygame.time.get_ticks() - ticks
        ticks += ticks_delta
        if current_time is not None:
            if sample_rate is None:
                current_time += ticks_delta / 1000.0 * time_multiplier
            else:
                current_time += 1.0 / sample_rate

        update(states)

        if current_time < start_time:
            continue
        if end_time is not None and current_time >= end_time:
            break

        display()

        while True:
            event = pygame.event.poll()
            if not event:
                break
            if event.type == QUIT:
                quit = True
            elif event.type == KEYDOWN or event.type == KEYUP:
                handle_key(event.key, event.type)

def parse_time(s):
    t = 0.0
    parts = s.split(":")
    parts.reverse()
    if len(parts) == 0 or len(parts) > 3:
        raise ValueError()
    if len(parts) > 0:
        t += float(parts[0])
    if len(parts) > 1:
        t += 60 * int(parts[1])
    if len(parts) > 2:
        t += 60 * 60 * int(parts[2])
    return t

def print_usage():
    print """\
Usage: %s [-h] [-m] [--start <time>] [--end <time>] <file name>

  -h, --help   Show this help.
  -m, --movie  Write raw RGB frames (suitable for encoding) to stdout.
  -r, --rate   Output frames at this rate (defaults to real time or 29.97 in
                 movie mode.
  -x           Width of the window.
  -y           Height of the window.
      --start [[hh:]mm:]ss.ss  Start playing at a certain time.
      --end [[hh:]mm:]ss.ss    End playing at a certain time.
""" % sys.argv[0]

def main():
    global screen_width, screen_height, sample_rate, start_time, end_time, movie_mode

    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:], "hmr:x:y:", ["help", "movie", "rate=", "start=", "end="])
    except getopt.GetoptError:
        print_usage()
        sys.exit(1)
    for o, a in opts:
        if o in ("-h", "--help"):
            print_usage()
            sys.exit()
        elif o in ("-m", "--movie"):
            movie_mode = True
        elif o in ("-r", "--rate"):
            sample_rate = float(a)
        elif o in ("-x"):
            screen_width = int(a)
        elif o in ("-y"):
            screen_height = int(a)
        elif o in ("--start"):
            start_time = parse_time(a)
        elif o in ("--end"):
            end_time = parse_time(a)
    if len(args) != 1:
        print_usage()
        sys.exit(1)
    file_name = args[0]

    file = open(file_name)
    try:
        run(file)
    finally:
        file.close()

if __name__ == "__main__":
    main()
